#include <amxmodx>
#include <fakemeta>
#include <engine>
#include <fun>
#include <hamsandwich>
#include <xs>
#include <orpheu>
#include <orpheu_memory>
#include <orpheu_advanced>

#if !defined _orpheu_included
	#assert "orpheu.inc library required ! Get it from https://forums.alliedmods.net/showthread.php?t=116393"
#endif

#if !defined _orpheu_memory_included
	#assert "orpheu_memory.inc library required ! Get it from https://forums.alliedmods.net/showthread.php?t=116393"
#endif

#if !defined _orpheu_advanced_included
	#assert "orpheu_advanced.inc library required ! Get it from https://forums.alliedmods.net/showthread.php?t=116393"
#endif

#define PluginName    "Grenade Drop & Default Ammo Control"
#define PluginVersion "0.1"
#define PluginAuthor  "HamletEagle"

#define MAX_ITEM_TYPES 6
#define MAX_AMMO_SLOTS 32

#define GetItemInfo(%0,%1)    (OrpheuMemoryGetAtAddress(ItemInfoArrayBaseAddr + %0 * MemberSize + %1 * StructureSize, MemoryDataName))
#define SetItemInfo(%0,%1,%2) (OrpheuMemorySetAtAddress(ItemInfoArrayBaseAddr + %0 * MemberSize + %1 * StructureSize, MemoryDataName, 1, %2))

#if AMXX_VERSION_NUM < 183

#define INT_BYTES        4 
#define BYTE_BITS        8 

stock bool:get_pdata_bool(ent, charbased_offset, intbase_linuxdiff = 5) 
{ 
    return !!( get_pdata_int(ent, charbased_offset / INT_BYTES, intbase_linuxdiff) & (0xFF<<((charbased_offset % INT_BYTES) * BYTE_BITS)) ) 
} 
#endif


enum _:Grenades
{
	FlashBang,
	HEGrenade,
	SmokeGrenade
}

enum
{
	PRIMARY_WEAPON_SLOT = 1,
	PISTOL_SLOT,
	KNIFE_SLOT,
	GRENADE_SLOT,
	C4_SLOT
}

enum _:Functions
{
	OrpheuFunction:InstallGamerules,
	OrpheuFunction:CreateNamedEntity,
	OrpheuFunction:Cvar_DirectSet,
	OrpheuFunction:PackWeapon,
	OrpheuFunction:Kill,
	OrpheuFunction:GetNextBestWeapon,
	OrpheuFunction:AmmoInventory,
	OrpheuFunction:DropPlayerItem
}

new const MemoryDataName        [] = "int"
new const WeaponBoxClassName    [] = "weaponbox"

new const m_rgAmmo_CBasePlayer        [MAX_AMMO_SLOTS] = {376, 377, ...}
new const m_rgpPlayerItems_CBasePlayer[MAX_ITEM_TYPES] = {367, 368, ...}
new const m_rgpPlayerItems_CWeaponBox [MAX_ITEM_TYPES] = {34 , 35 , ...}

new const GrenadeClassNames[Grenades][] =
{
	"weapon_flashbang",
	"weapon_hegrenade",
	"weapon_smokegrenade"
}

const m_pActiveItem	     = 373 //active item
const m_bIsVip           = 837 //true if a player is VIP
const m_iType   		 = 34  //armoury_entity type
const m_pNext       	 = 42  //next linked weapon 
const m_iId	       	     = 43  //weapon index 
const m_iPrimaryAmmoType = 49  //primary ammo index into players m_rgAmmo[]
const m_pfnThink    	 = 4   //think function for entity
const XoPlayer      	 = 5   //player    linux offset = windows offset + 5
const XoCBasePlayerItem  = 4   //item      linux offset = windows offset + 4
const XoCWeaponBox		 = 4   //weaponbox linux offset = windows offset + 4

//Grenades buy slot
const FlashBangBuySlot    = 3
const HEGrenadeBuySlot    = 4
const SmokeGrenadeBuySlot = 5

//Default ammo for grenades
const FlashBangDefaultAmmo 	  = 2
const HEGrenadeDefaultAmmo 	  = 1
const SmokeGrenadeDefaultAmmo = 1

//Default ammo indexes for grenades. Since this doesn't change I've hardcoded them
const FlashBangAmmoIndex    = 11
const HEGrenadeAmmoIndex    = 12
const SmokeGrenadeAmmoIndex = 13

//Grenade indexes when packed into armoury_entity
const FlashBangArmoury    = 14
const HEGrenadeArmoury    = 15
const SmokeGrenadeArmoury = 18

//For ItemInfoArray
const MemberSize     = 4
const MaxAmmo1Member = 3
const StructureSize  = 44

new ItemInfoArrayBaseAddr
new CWeaponBoxKillFuncAddress
new WeaponBoxClassNameOffset
new CvarDropGrenades
new g_pGameRules
new gmsgTextMsg
  
new SlotToWeaponIndex    [Grenades]
new GrenadesDefaultAmmo  [Grenades]
new CvarMaxGrenades      [Grenades]
new CvarMaxGrenadesValues[Grenades]
new OrpheuFunction:HandleGameFunction[Functions]
new OrpheuHook:HandleDropPlayerItemHook

public plugin_precache()
{
	HandleGameFunction[InstallGamerules] = OrpheuGetFunction("InstallGameRules")
	OrpheuRegisterHook(HandleGameFunction[InstallGamerules], "OnInstallGameRules", OrpheuHookPost)
}

public OnInstallGameRules()
{
	g_pGameRules = OrpheuGetReturn()    
	HandleGameFunction[GetNextBestWeapon] = OrpheuGetFunctionFromObject(g_pGameRules, "GetNextBestWeapon", "CGameRules")
}

public plugin_init()
{
	register_plugin
	(
		.plugin_name = PluginName,
		.version     = PluginVersion,
		.author      = PluginAuthor
	)

	CvarMaxGrenades[FlashBang   ] = register_cvar("max_flashbangs"   , "2")
	CvarMaxGrenades[HEGrenade   ] = register_cvar("max_hegrenades"   , "1")
	CvarMaxGrenades[SmokeGrenade] = register_cvar("max_smokegrenades", "1")
	CvarDropGrenades              = register_cvar("drop_grenades"    , "1")

	new const CBasePlayer[] = "CBasePlayer"
	new const CWeaponBox [] = "CWeaponBox"

	HandleGameFunction[Cvar_DirectSet   ] = OrpheuGetFunction("Cvar_DirectSet"             )
	HandleGameFunction[CreateNamedEntity] = OrpheuGetFunction("CREATE_NAMED_ENTITY"        )
	HandleGameFunction[AmmoInventory    ] = OrpheuGetFunction("AmmoInventory" , CBasePlayer)
	HandleGameFunction[DropPlayerItem   ] = OrpheuGetFunction("DropPlayerItem", CBasePlayer)
	HandleGameFunction[Kill             ] = OrpheuGetFunction("Kill"          , CWeaponBox )
	HandleGameFunction[PackWeapon       ] = OrpheuGetFunction("PackWeapon"    , CWeaponBox )
	
	OrpheuRegisterHook(HandleGameFunction[Cvar_DirectSet], "OnCvar_DirectSet"          , OrpheuHookPre)
	OrpheuRegisterHook(HandleGameFunction[AmmoInventory ], "CBasePlayer_AmmoInventory" , OrpheuHookPre)
	
	if(get_pcvar_num(CvarDropGrenades))
	{
		HandleDropPlayerItemHook = OrpheuRegisterHook(HandleGameFunction[DropPlayerItem], "CBasePlayer_DropPlayerItem", OrpheuHookPre)
	}
	
	RegisterHam(Ham_Touch, WeaponBoxClassName, "CWeaponBox_Touch", false)
	
	/*
		| When a weaponbox entity is created, game sets think to CWeaponBox::Kill
		| And pev_nextthink to gametime() + 300.0
		| This creates a timer that remove the box after 300 seconds
		| Kill() basically get all items from weaponbox and remove them + remove the box
		| Calling think on a custom weaponbox will do nothing, because game does not know what to do with it when it thinks
		| Here we retrieve CWeaponBox::Kill function address so we can set later m_pfnThink offset and let game handle the weaponbox.
		| Another way would be to create a bot, give him a weapon, force drop, hook Touch and retrieve the address.
	*/
	CWeaponBoxKillFuncAddress = OrpheuGetFunctionAddress(HandleGameFunction[Kill])
	
	//Thanks Arkshine for finding the real address at which ItemInfoArray starts
	ItemInfoArrayBaseAddr = OrpheuMemoryGet("ItemInfoArray") - (is_linux_server() ? 0 : 24)
	
	SlotToWeaponIndex[FlashBang   ] = CSW_FLASHBANG
	SlotToWeaponIndex[HEGrenade   ] = CSW_HEGRENADE
	SlotToWeaponIndex[SmokeGrenade] = CSW_SMOKEGRENADE
	
	GrenadesDefaultAmmo[FlashBang   ] = FlashBangDefaultAmmo
	GrenadesDefaultAmmo[HEGrenade   ] = HEGrenadeDefaultAmmo
	GrenadesDefaultAmmo[SmokeGrenade] = SmokeGrenadeDefaultAmmo

	WeaponBoxClassNameOffset = engfunc(EngFunc_AllocString, WeaponBoxClassName)
	
	for(new i; i < sizeof CvarMaxGrenades; i++)
	{
		CvarMaxGrenadesValues[i] = get_pcvar_num(CvarMaxGrenades[i])
		PatchMemory(CvarMaxGrenadesValues[i], SlotToWeaponIndex[i])
	}
}

public plugin_end()
{
	UnpatchMemory()
}

public plugin_pause()
{
	UnpatchMemory()
}

public plugin_unpause()
{
	for(new i; i < sizeof CvarMaxGrenades; i++)
	{
		PatchMemory(CvarMaxGrenadesValues[i], SlotToWeaponIndex[i])
	}
}

/*
	| Patching only one time is enough, because nothing alters ItemInfoArray.
	| But given that the purpose of the plugin is to allow dynamic changes we call again PatchMemory when a cvar is changed.
	| Cvar_DirectSet is the function that actually sets cvar values.
*/
public OnCvar_DirectSet(const cvarPointer, const cvarValue[])
{
	new cvarNewValue = str_to_num(cvarValue)
	new cvarOldValue
	
	if(cvarPointer == CvarDropGrenades)
	{
		cvarOldValue = get_pcvar_num(CvarDropGrenades)
		if(cvarNewValue != cvarOldValue)
		{
			if(cvarNewValue)
			{
				HandleDropPlayerItemHook = OrpheuRegisterHook(HandleGameFunction[DropPlayerItem], "CBasePlayer_DropPlayerItem", OrpheuHookPre)
			}
			else 
			{
				OrpheuUnregisterHook(HandleDropPlayerItemHook)
			}
		}
	}
	else 
	{
		for(new i; i < sizeof CvarMaxGrenades; i++)
		{
			if(cvarPointer == CvarMaxGrenades[i])
			{
				cvarOldValue = CvarMaxGrenadesValues[i]
				if(cvarNewValue != cvarOldValue)
				{
					CvarMaxGrenadesValues[i] = cvarNewValue
					PatchMemory(cvarNewValue, SlotToWeaponIndex[i])
					break
				}
			}
		}
	}
}

/*
	| Called in order to drop player item(also from "drop" command)
	| If itemName is null then player wants to drop active item.
	| Inside game function there is a check with CanDrop() which simply return true/false depending on item 
	| We could simply hook this function and set it's return value to true, butgame is not designed to drop grenades 
	| so such situation is not properly handled.
	| For example, if you have more grenades you won't drop just one, you will drop all of them. 
	| Considering default game behaviour, this is only an issue with flashbang, but given that this plugin also change max grenade ammout 
	| it was a must to redo partially the function.
*/
public OrpheuHookReturn:CBasePlayer_DropPlayerItem(id, const itemName[])
{
	new weaponEntity = FM_NULLENT, activeItem = FM_NULLENT
	new weaponClassName[32]
	
	if(is_user_alive(id))
	{
		weaponEntity = get_pdata_cbase(id, m_rgpPlayerItems_CBasePlayer[GRENADE_SLOT], XoPlayer)
		activeItem   = get_pdata_cbase(id, m_pActiveItem, XoPlayer)
		
		while(pev_valid(weaponEntity))
		{
			if(itemName[0] != EOS)
			{
				pev(weaponEntity, pev_classname, weaponClassName, charsmax(weaponClassName))
				if(equal(itemName, weaponClassName))
				{
					break
				}
			}
			else
			{
				if(weaponEntity == activeItem)
				{
					break
				}
			}
			
			weaponEntity = get_pdata_cbase(weaponEntity, m_pNext, XoCBasePlayerItem)
		}
		
		//Player wants to drop a grenade, we handle this. For other items, let game do this job.
		//Basically, we redo some of the code from SDK.
		if(pev_valid(weaponEntity))
		{
			if(get_pdata_bool(id, m_bIsVip, XoPlayer))
			{
				if(!gmsgTextMsg)
				{
					gmsgTextMsg = get_user_msgid("TextMsg")
				}
				
				message_begin(MSG_ONE, gmsgTextMsg, .player = id)
				{
					write_byte(print_center)
					write_string("#Weapon_Cannot_Be_Dropped")
					message_end()
				}

				return OrpheuSupercede
			}
			
			new ammoIndex = ExecuteHam(Ham_Item_PrimaryAmmoIndex, weaponEntity)
			if(ammoIndex != -1)
			{
				new userBackpackAmmo = get_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], XoPlayer)
				
				if(userBackpackAmmo)
				{
					new weaponIndex = get_pdata_int(weaponEntity, m_iId, XoCBasePlayerItem)
					new dummyEntity = FM_NULLENT
					new weaponBoxEntity 
					
					if(weaponClassName[0] == EOS)
					{
						//If we are here then player wants to drop active item, get it's classname
						pev(weaponEntity, pev_classname, weaponClassName, charsmax(weaponClassName))
					}
					
					do 
					{
						/*
							| Create a new entity, this will be packed into the box instead of player grenade
							| If we pack player grenade, then item will be removed, even if he has more grenades in inventory
							| Probably better to create one only if userBackpackAmmo > 1, else pack default grenade
							| Something to change later
						*/
						dummyEntity = create_entity(weaponClassName)
					}
					while(!pev_valid(dummyEntity))

					set_pev(dummyEntity, pev_spawnflags, pev(dummyEntity, pev_spawnflags) | SF_NORESPAWN)
					set_pdata_int(dummyEntity, m_iPrimaryAmmoType, ammoIndex, XoCBasePlayerItem)
					
					ExecuteHam(Ham_Spawn, dummyEntity)
					
					/*
						| Due to optimizations entities created by game are cached and searching functions use the hash table to find them.
						| weaponbox entities should be removed when round ends, but if you use default amxx functions like create_entity() 
						| entity won't be insterted in hash table => game will ignore it and default behaviour will be breaked.
						| In CS, there is a custom function called CREATE_NAMED_ENTITY which calls both AddEntityHashValue and pfnCreateNamedEntity().
					*/
					#if AMXX_VERSION_NUM < 183
						weaponBoxEntity = OrpheuCallSuper(HandleGameFunction[CreateNamedEntity], WeaponBoxClassNameOffset)
					#else 
						weaponBoxEntity = cs_create_entity(WeaponBoxClassName)
					#endif
					
					new Float:playerOrigin[3]; pev(id, pev_origin, playerOrigin)
					new Float:playerAngles[3]; pev(id, pev_angles, playerAngles)
				
					new Float:forwardDirection[3]
					new Float:forwardDirection2[3]
					engfunc(EngFunc_MakeVectors, playerAngles)
					global_get(glb_v_forward, forwardDirection)
					
					xs_vec_mul_scalar(forwardDirection, 10.0, forwardDirection)
					xs_vec_add(playerOrigin, forwardDirection, playerOrigin)
					
					set_pev(weaponBoxEntity, pev_origin, playerOrigin)
					ExecuteHam(Ham_Spawn, weaponBoxEntity)
					
					playerAngles[0] = 0.0
					playerAngles[2] = 0.0
					set_pev(weaponBoxEntity, pev_angles, playerAngles)

					set_pdata_int(weaponBoxEntity, m_pfnThink, CWeaponBoxKillFuncAddress, 0)
					set_pev(weaponBoxEntity, pev_nextthink, get_gametime() + 300)
					
					set_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], userBackpackAmmo - 1, XoPlayer)
						
					if(userBackpackAmmo - 1 <= 0)
					{
						OrpheuCallSuper(HandleGameFunction[GetNextBestWeapon], g_pGameRules, id, weaponEntity)
							
						if(activeItem == weaponEntity)
						{
							/*
								| If player drop active item, then we can Holster() it and let game do some of the function calls for us.
								| Grenades have a special holster.
								| When ammo is 0 it cleans weapon hud and calls RemovePlayerItem().
								| Third param is skiplocal.
							*/
							ExecuteHamB(Ham_Item_Holster, weaponEntity, 0)
						}
						else 
						{
							/*
								| There is nothing to holster, we need to do above manually
							*/
							new userWeapons = pev(id, pev_weapons)
							userWeapons &= ~ (1 << weaponIndex)
							set_pev(id, pev_weapons, userWeapons)
							ExecuteHamB(Ham_RemovePlayerItem, id, weaponEntity)
						}
					}
					
					OrpheuCallSuper(HandleGameFunction[PackWeapon], weaponBoxEntity, dummyEntity)
					
					xs_vec_copy(forwardDirection, forwardDirection2)
					xs_vec_mul_scalar(forwardDirection , 30.0, forwardDirection)
					xs_vec_mul_scalar(forwardDirection2, 10.0, forwardDirection2)
					xs_vec_add(forwardDirection, forwardDirection2, forwardDirection)
					set_pev(weaponBoxEntity, pev_velocity, forwardDirection)
					
					new modelName[32]
					GetCSModelName(weaponIndex, modelName, charsmax(modelName))
					engfunc(EngFunc_SetModel, weaponBoxEntity, modelName)
				}
				
				return OrpheuSupercede
			}
		}
	}
	return OrpheuIgnored
}

/*
	| Called from CArmoury::ArmouryTouch() and BuyItem()
	| It is used to check if player can get another grenade.
	| Since default values are hardcoded(2/1/1) and redoing the function will be painful
	| The best option is to alter the return value of AmmoInventory
*/
public OrpheuHookReturn:CBasePlayer_AmmoInventory(const id, const ammoIndex)
{
	if(FlashBangAmmoIndex <= ammoIndex <= SmokeGrenadeAmmoIndex)
	{
		new grenadeType = ammoIndex - 11
		new maxGrenadeAmmount = CvarMaxGrenadesValues[grenadeType]
		
		if(maxGrenadeAmmount != GrenadesDefaultAmmo[grenadeType]) //if cvars are set to default values, don't bother doing anything
		{
			new grenadesCount = get_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], XoPlayer)
			new returnValue = -1
			
			if(grenadesCount < maxGrenadeAmmount)
			{
				returnValue = 0
			}
			else 
			{
				if(maxGrenadeAmmount < GrenadesDefaultAmmo[grenadeType])
				{	
					/*
						| Above is handled the case when cvar values are bigger than default values 
						| Here we take care of lower values.
					*/
					returnValue = GrenadesDefaultAmmo[grenadeType] + 1
				}
			}
			
			if(returnValue != -1)
			{
				OrpheuSetReturn(returnValue)
				return OrpheuOverride
			}
		}
	}
	return OrpheuIgnored
}

/*
	| As stated before, game was not coded to allow grenade drop and it gets pretty bugged when we change that.
	| For example, if you pickup one grenade you will receive two, because a new grenade entity is spawned.
	| Also, max ammount of grenade is hardcoded, the only solution is to handle manually grenade touch case.
*/
public CWeaponBox_Touch(weaponBoxEntity, id)
{	
	if(pev_valid(weaponBoxEntity))
	{
		if(!(pev(weaponBoxEntity, pev_flags) & FL_ONGROUND))
		{
			return HAM_IGNORED
		}
		
		if(!is_user_alive(id))
		{
			return HAM_IGNORED
		}
		
		new weaponEntity = get_pdata_cbase(weaponBoxEntity, m_rgpPlayerItems_CWeaponBox[GRENADE_SLOT], XoCWeaponBox)
		new ammoIndex, playerGrenades, maxGrenades, grenadeName[32], weaponIndex
		
		if(pev_valid(weaponEntity))
		{
			do
			{
				ammoIndex = ExecuteHam(Ham_Item_PrimaryAmmoIndex, weaponEntity)
				if(ammoIndex != -1)
				{
					playerGrenades = get_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], XoPlayer)
					maxGrenades    = 0

					weaponIndex = get_pdata_int(weaponEntity, m_iId, XoCBasePlayerItem)
					switch(weaponIndex)
					{
						case CSW_HEGRENADE:
						{
							copy(grenadeName, charsmax(grenadeName), GrenadeClassNames[HEGrenade])
							maxGrenades = CvarMaxGrenadesValues[HEGrenade]
						}
						case CSW_SMOKEGRENADE:
						{
							copy(grenadeName, charsmax(grenadeName), GrenadeClassNames[SmokeGrenade])
							maxGrenades = CvarMaxGrenadesValues[SmokeGrenade]
						}
						case CSW_FLASHBANG:
						{
							copy(grenadeName, charsmax(grenadeName), GrenadeClassNames[FlashBang])
							maxGrenades = CvarMaxGrenadesValues[FlashBang]
						}
					}
					
					if(playerGrenades < maxGrenades && grenadeName[0] != EOS)
					{
						give_item(id, grenadeName)
						
						//Unlink the weapon from the box
						weaponEntity = get_pdata_cbase(weaponEntity, m_pNext, XoCWeaponBox)
						set_pdata_cbase(weaponBoxEntity, m_rgpPlayerItems_CWeaponBox[GRENADE_SLOT], weaponEntity)

						continue
					}
				}
				
				weaponEntity = get_pdata_cbase(weaponEntity, m_pNext, XoCWeaponBox)
			}
			while(pev_valid(weaponEntity))
			
			return HAM_SUPERCEDE
		}
	}
	
	return HAM_IGNORED
}

/*
	| Retrieve model to set on the box.
	| In CS, SetModel() is fired two times on a weaponbox.
	| First time, for w_weaponbox.mdl, then it set the model according to the packed weapon.
*/
GetCSModelName(const itemId, modelName[], const Size)
{
	switch(itemId)
	{
		case CSW_HEGRENADE   :	copy(modelName, Size, "models/w_hegrenade.mdl"   )
		case CSW_SMOKEGRENADE:	copy(modelName, Size, "models/w_smokegrenade.mdl")
		case CSW_FLASHBANG   :	copy(modelName, Size, "models/w_flashbang.mdl"   )
	}
}

PatchMemory(const maxGrenadeAmmount, const weaponIndex)
{
	if(maxGrenadeAmmount != GetItemInfo(MaxAmmo1Member, weaponIndex))
	{
		SetItemInfo(MaxAmmo1Member, weaponIndex, maxGrenadeAmmount)
	}
}

UnpatchMemory()
{
	for(new i; i < sizeof SlotToWeaponIndex; i++)
	{
		if(GetItemInfo(MaxAmmo1Member, SlotToWeaponIndex[i]) != GrenadesDefaultAmmo[i])
		{
			SetItemInfo(MaxAmmo1Member, SlotToWeaponIndex[i], GrenadesDefaultAmmo[i])
		}
	}
}